from sklearn.neighbors import NearestNeighbors, LocalOutlierFactor
import pandas as pd
import numpy as np
from ..utils import cat_encode, df_combine, df_separate


def get_WEAP(synth, key, target):
    """Get the Within Equivalence class Attribution Probabilities WEAP

    For each record in the synthetic dataset, this function returns the
    proportion across the whole dataset that these `key` values are matched
    with this `target` value.

    Parameters
    ----------
    synth : pandas dataframe
        Dataframe containing the synthetic data.
    key : list
        List of features in `synth` to use as the key.
    target : str or list of str
        Feature to use as the target.

    Returns
    -------
    pandas.Series
        A series object containing the WEAP scores for each record in `synth`.

    Notes
    -----
    This function is intended to only be used within `TCAP()` to determine
    which synthetic records are most likely to pose an attribution risk.
    """
    synth = synth.copy()
    key_target_vc = synth.value_counts(subset=key + target).reset_index()
    key_target_vc.columns = key + target + ['key_target_frequency']
    key_vc = synth.value_counts(key).reset_index()
    key_vc.columns = key + ['key_frequency']

    synth = synth.merge(key_target_vc)
    synth = synth.merge(key_vc)

    return synth['key_target_frequency']/synth['key_frequency']


def TCAP(real, synth, key, target):
    """ Target Correct Attribution Probability TCAP

    This privacy metric calculates the average chance that the key-target
    pairings in the `synth` dataset reveal the true key-target pairings in the
    original, `real` dataset.

    Parameters
    ----------
    real : pandas dataframe
        Dataframe containing the real data.
    synth : pandas dataframe
        Dataframe containing the synthetic data.
    key : list
        List of features in `synth` to use as the key.
    target : str or list of str
        Feature to use as the target.

    Returns
    -------
    TCAP : float
        The average TCAP across the dataset.

    Notes
    -----
    This metric provides an estimate of how well an intruder could infer
    attributes of groups in the real dataset by studying the synthetic. The
    choices for `key` and `target` will vary depending on the dataset in
    question but we would suggest the `key` features are those that could be
    readily available to an outsider and the `target` feature is one we
    wouldn't want them finding out, such as a protected characteristic.

    This method only works with categorical data, so binning of continuous data
    may be required.
    """
    if not isinstance(target, list):
        target = [target]
    WEAP_scores = get_WEAP(synth, key, target)
    if sum(WEAP_scores == 1) == 0:
        return 0
    synth_reduced = synth[WEAP_scores == 1][key+target]
    synth_reduced.drop_duplicates(inplace=True)
    synth_reduced.rename(
        columns={target[0]: target[0]+'_synthguess'}, inplace=True)

    combined = real.merge(synth_reduced, how='left', on=key)
    target_matches = combined[target[0]] == combined[target[0]+'_synthguess']

    return sum(target_matches)/len(target_matches)


def find_outliers(data, outlier_factor_threshold):
    """Find Outliers

    This function returns whether each row in `data` can be considered an
    outlier.

    Parameters
    ----------
    data : pandas dataframe
    outlier_factor_threshold : float
        Float influencing classification of ouliers. Increasing this threshold
        means that fewer points are considered outliers.

    Returns
    -------
    outlier_bool : list of bool
        List indicating which rows of `data` are outliers.

    Notes
    -----
    Most inliers will have an outlier factor of less than one, however there
    are no clear rules that determine when a data point is an outlier. This
    is likely to vary from dataset to dataset and, as such, we recommend
    tuning `outlier_factor_threshold` to suit.
    """
    lof = LocalOutlierFactor(n_neighbors=5)
    lof.fit_predict(data)
    outlier_factor = -lof.negative_outlier_factor_
    outlier_bool = outlier_factor < outlier_factor_threshold
    return outlier_bool


def min_NN_dist(real, synth, feats=None, real_outliers_only=True,
                outlier_factor_threshold=2):
    """ Minimum Nearest Neighbour distance

    This privacy metric returns the smallest distance between any point in
    the `real` dataset and any point in the `synth` dataset. There is an
    option to only consider the outliers in the real dataset as these perhaps
    pose more of a privacy concern.

    Parameters
    ----------
    real : pandas dataframe
        Dataframe containing the real data.
    synth : pandas dataframe
        Dataframe containing the synthetic data.
    feats: str or list of str, optional
        Features to use. By default all features are used.
    real_outliers_only : bool (default True)
        Boolean indicating whether to filter out inliers (default) or not.
    outlier_factor_threshold : float (default 2)
        Float influencing classification of ouliers. Increase to include
        fewer real points in nearest neighbour calculations.

    Returns
    -------
    min_dist : float
        Minimum manhattan distance between `real` and `synth` data.

    Notes
    -----
    This privacy metric provides an insight into whether the synthetic dataset
    is too similar to the real dataset. It does this by calculating the
    minimum distance between the real records and the synthetic records.

    This metric assumes that categorical data is ordinal during distance
    calculations, or that it has already been suitably one-hot-encoded.
    """
    combined = df_combine(real, synth, feats=feats)
    combined_recode, _ = cat_encode(combined, return_all=True)
    real, synth = df_separate(combined_recode, source_col_name='source',
                              source_val_real=0, source_val_synth=1)
    if real_outliers_only:
        outlier_bool = find_outliers(real, outlier_factor_threshold)
        real = real[outlier_bool]
    near_neigh = NearestNeighbors(n_neighbors=1, radius=100, p=1)
    near_neigh.fit(real)
    dist, ind = near_neigh.kneighbors(synth, return_distance=True)
    return int(min(dist))


def sample_overlap_score(real, synth, feats=None, sample_size=0.2, runs=5,
                         score_type='unique'):
    """ Return percentage of overlap between real and synth data.

    Samples from both the real and synthetic datasets are compared for
    similarity. This similarity, or overlap score, is based on the
    exact matches of real data records within the synthetic data.

    Parameters
    ----------
    real: pandas.DataFrame
        DataFrame containing the real data.
    synth: pandas.DataFrame
        DataFrame containing the synthetic data.
    feats: str or list of str, optional.
        The features that will be used to match records. By
        default all features will be used.
    sample_size: float or int, optional
        The ratio (if `sample_size` between 0 and 1) or count
        (`sample_size` > 1) of records to sample. Default is 0.2 or 20%.
    runs: int, optional
        The number of times to compute the score. Total score is averaged
        across runs.
    score_type: {"unique"|"sample"}
        Method used for calculating the overlap score. If "unique", the
        default, the score is the percentage of unique records in the real
        sample that have a match within the synth data. If "sample" the
        score is the percentage of all records within the real sample that
        have a match within the synth sample.

    Returns
    -------
    float:
        Overlap score between `real` and `synth`
    """
    if isinstance(feats, pd.Index):
        feats = feats
    elif isinstance(feats, str):
        feats = [feats]
    else:
        feats = feats or real.columns.to_list()
    if 0 <= sample_size <= 1:
        nsamples = int(real.shape[0] * sample_size)
    else:
        if sample_size > real.shape[0]:
            nsamples = real.shape[0]
        else:
            nsamples = sample_size
    scores = []
    for r in range(runs):
        sample_real = real[feats].sample(nsamples).assign(real_count=1) \
            .groupby(feats).count().reset_index()
        sample_synth = synth[feats].sample(nsamples).assign(synth_count=1) \
            .groupby(feats).count().reset_index()
        duplicates = sample_real.merge(sample_synth, how='left', on=feats,
                                       suffixes=('_real', '_synth'),
                                       indicator='_match')
        if score_type == 'unique':
            score = duplicates._match.value_counts(normalize=True).both
        elif score_type == 'sample':
            score = duplicates[duplicates._match ==
                               'both'].real_count.sum() / nsamples
        scores.append(score)
    return np.mean(scores)


if __name__ == "__main__":
    pass



from scipy.stats import (
    entropy, kruskal, ks_2samp, mannwhitneyu, wasserstein_distance)
from scipy.stats import wilcoxon as scipy_wc
from scipy.spatial.distance import jensenshannon
import pandas as pd
import numpy as np

from ..utils import feature_density_diff


def kolmogorov_smirnov(real, synth, feature, **kwargs):
    """ Distance: Kolmogorov-Smirnov

    The Kolmogorov-Smirnov metric calculates the maximum difference between the
    cumulative distribution functions of the `feature` in the `real` and
    `synth` datasets.

    Parameters
    ----------
    real : pandas dataframe
        Dataframe containing the real data.
    synth : pandas dataframe
        Dataframe containing the synthetic data.
    feature : str
        String indicating the feature of the datasets to compare. This must be
        a continuous variable.
    **kwargs : dict
        Keyword arguments.

    Returns
    -------
    statistic : float
        Kolmogorov-Smirnov statistic
    pvalue : float
        Two-tailed p-value

    See Also
    --------
    scipy.stats.ks_2samp

    Notes
    -----
    This is a wrapper function for `scipy.stats.ks_2samp`, which tests whether
    two samples are drawn from the same distribution by calculating the maximum
    difference between their cumulative distribution functions.

    If the returned statistic is small or the p-value is high, then we cannot
    reject the hypothesis that the distributions are the same.

    This approach is only defined if the feature is continuous. The SciPy
    documentation further suggests this method works best when one of the
    samples has a size of only a few thousand.

    Examples
    --------
    >>> import pandas as pd
    >>> real_data = pd.DataFrame(get_real(500),
    ...                          columns = ['feat1', 'feat2', 'feat3'])
    >>> synth_data = pd.DataFrame(get_synth(500),
    ...                           columns = ['feat1', 'feat2', 'feat3'])

    The first feature appears to come from the same distribution in both
    datasets.

    >>> kolmogorov_smirnov(real_data, synth_data, 'feat1')
    KstestResult(statistic=0.062, pvalue=0.2919248807417811) # random

    The second feature appears to come from different distributions in the
    datasets.

    >>> kolmogorov_smirnov(real_data, synth_data, 'feat2')
    KstestResult(statistic=0.274, pvalue=6.383314923658339e-17) # random
    """
    return ks_2samp(real[feature], synth[feature], **kwargs)


def wasserstein(real, synth, feature, **kwargs):
    """ Distance: Wasserstein

    The Wasserstein distance, or Earth Mover's distance, can be thought of as
    calculating the amount of "work" required to move from the distribution of
    the synthetic data to the distribution of the real data.

    Parameters
    ----------
    real : pandas dataframe
        Dataframe containing the real data.
    synth : pandas dataframe
        Dataframe containing the synthetic data.
    feature : str
        String indicating the feature of the datasets to compare. This must be
        a continuous variable.
    **kwargs : dict
        Keyword arguments.

    Returns
    -------
    distance : float
        The computed distance between the distributions.

    See Also
    --------
    scipy.stats.wasserstein_distance

    Notes
    -----
    This is a wrapper function for `scipy.stats.wasserstein_distance`.

    Computationally, we can find the Wasserstein distance by calculating the
    area between the cumulative distribution functions for the two
    distributions.

    If :math:`s` is the `synth` distribution of the `feature`, :math:`r` is the
    `real` distribution of the `feature` and :math:`R` and :math:`S` are their
    respective cumulative distribution functions, then

    .. math::

        wasserstein(s, r) = \\int_{-\\infty}^{+\\infty} |S-R|

    The distance is zero if the distributions are identical, and increases as
    they become less alike. This method is therefore good for comparing
    multiple synthetic datasets, or features within a dataset, to see which is
    closest to the real. However, as this is not a test, there is no threshold
    distance below which we can claim the distributions are statistically the
    same.

    Examples
    --------
    >>> import pandas as pd
    >>> real_data = pd.DataFrame(get_real(500),
    ...                          columns = ['feat1', 'feat2', 'feat3'])
    >>> synth_data = pd.DataFrame(get_synth(500),
    ...                           columns = ['feat1', 'feat2', 'feat3'])

    The first feature appears to be more similar than the second across
    datasets.

    >>> wasserstein(real_data, synth_data, 'feat1')
    0.0688192355094602 # random
    >>> wasserstein(real_data, synth_data, 'feat2')
    0.8172329918412307 # random

    """
    return wasserstein_distance(real[feature], synth[feature], **kwargs)


def jensen_shannon_distance(real, synth, feature, bins='auto', **kwargs):
    """ Distance: Jensen-Shannon

    The Jensen-Shannon distance describes the difference between the `real` and
    `synth` distributions of the `feature` in terms of entropy. It is the
    square root of Jensen-Shannon divergence. It measures the distance between
    probabilities so the data are first discretised into bins.

    Parameters
    ----------
    real : pandas dataframe
        Dataframe containing the real data.
    synth : pandas dataframe
        Dataframe containing the synthetic data.
    feature : str
        String indicating the feature of the datasets to compare. This must be
        a continuous variable.
    bins : int or str
        Number of bins to use when discretising data, if `None` data will be
        treated as categorical. If string chosen method will be used to
        calculate optimal bin width, 'auto' is the default. See
        numpy.histogram_bin_edges for further options.
    **kwargs : dict
        Keyword arguments.

    Returns
    -------
    distance : double
        The computed distance between the distributions.

    See Also
    --------
    divergence.jensen_shannon_divergence
    scipy.spatial.distance.jensenshannon

    Notes
    -----
    This is a wrapper function for `scipy.spatial.distance.jensenshannon`.
    Since this function expects probability vectors the data is first
    discretised into evenly-spaced bins.

    We can think of the Jensen-Shannon distance as the amount of information,
    or entropy, encoded in the difference between the `real` and `synth`
    distributions of the `feature`.

    The distance is zero if the distributions are identical, and is bounded
    above by one if they are nothing alike. This method is therefore good
    for comparing multiple synthetic datasets, or features within a dataset,
    to see which is closest to the real. However, as this is not a test,
    there is no threshold distance below which we can claim the distributions
    are statistically the same.

    An optimal 'bins' value has not been suggested.

    Examples
    --------
    >>> import pandas as pd
    >>> real_data = pd.DataFrame(get_real(500),
    ...                          columns = ['feat1', 'feat2', 'feat3'])
    >>> synth_data = pd.DataFrame(get_synth(500),
    ...                           columns = ['feat1', 'feat2', 'feat3'])

    The first feature appears to be more similar than the second across
    datasets.

    >>> jensen_shannon_distance(real_data, synth_data, 'feat1', bins = 20)
    0.11006632967333475 # random
    >>> jensen_shannon_distance(real_data, synth_data, 'feat2', bins = 20)
    0.43556476029981644 # random

    """
    if bins is None:
        values = set(np.concatenate(
            [np.unique(real[feature]), np.unique(synth[feature])]))
        real_bincounts = [(real[feature].values == val).sum()
                          for val in values]
        synth_bincounts = [(synth[feature].values == val).sum()
                           for val in values]
    else:
        # assign bin edges based on joined data
        bin_edges = np.histogram_bin_edges(np.concatenate((real[feature],
                                                           synth[feature])),
                                           bins=bins)
        # count the number of points in each bin seperately
        real_bincounts = np.histogram(real[feature], bin_edges)[0]
        synth_bincounts = np.histogram(synth[feature], bin_edges)[0]
    return jensenshannon(real_bincounts, synth_bincounts, **kwargs)


def feature_density_diff_mae(real, synth, feats=None, bins=10):
    """ Calculate Mean Absolute Error of feature densities.

    For each feature the difference between the density across the bins
    within `real` and `synth` is calculated. Finally the MAE across all
    features and bins is calculated. A value close to 0 indicates that
    a similar distribution for `feats` is observed between the real and
    synthetic datasets.

    Parameters
    ----------
    real: pandas.DataFrame
        DataFrame containing the real data.
    synth: pandas.DataFrame
        DataFrame containing the sythetic data.
    feats: str or list of str, optional.
        The features that will be used to compute the densities. By
        default all features in `real` will be used.
    bins: str or int, optional
        Bins to use for computing the density. This value is passed
        to `numpy.histogram_bin_edges` so can be any value accepted by
        that function. The default setting of 10 uses 10 bins.

    Returns
    -------
    float:
        Mean Absolute Error of feature densities.
    """
    if isinstance(feats, pd.Index):
        feats = feats
    elif isinstance(feats, str):
        feats = [feats]
    else:
        feats = feats or real.columns
    diffs = [feature_density_diff(real, synth, f, bins)[0] for f in feats]
    return np.mean(np.abs(np.concatenate(diffs)))


def kullback_leibler(real, synth, feature, bins='auto', **kwargs):
    """ Divergence: Kullback-Leibler

    The Kullback-Leibler divergence describes how much the `real` distribution
    of the `feature` varies from the `synth` in terms of entropy. This is an
    assymmetric measure so does not describe the opposing variation. Since it
    measures the variation between probabilities, the data are first
    discretised into bins.


    Parameters
    ----------
    real : pandas dataframe
        Dataframe containing the real data.
    synth : pandas dataframe
        Dataframe containing the synthetic data.
    feature : str
        String indicating the feature of the datasets to compare. This must be
        a continuous variable.
    bins : int or str
        Number of bins to use when discretising data, if `None` data will be
        treated as categorical. If string chosen method will be used to
        calculate optimal bin width, 'auto' is the default. See
        numpy.histogram_bin_edges for further options.
    **kwargs : dict
        Keyword arguments.

    Returns
    -------
    D : float
        The calculated divergence.

    See Also
    --------
    scipy.stats.entropy

    Notes
    -----
    This is a wrapper function for `scipy.stats.entropy`. Since this function
    expects a probability vector, the data is first discretised into evenly-
    spaced bins.

    We can think of the Kullback-Leibler divergence as a measure of surprise
    we might expect seeing an example from the real data, relative to the
    distribution of the synthetic.

    The divergence is zero if the distributions are identical, and is bounded
    above by one if they are nothing alike. This method is therefore good
    for comparing multiple synthetic datasets, or features within a dataset,
    to see which is closest to the real. However, as this is not a test,
    there is no threshold distance below which we can claim the distributions
    are statistically the same.

    An optimal 'bins' value has not been suggested.

    BUG: returns `inf` if no real data falls in any one of the bins - dividing
    by zero error.

    Examples
    --------
    >>> import pandas as pd
    >>> real_data = pd.DataFrame(get_real(500),
                                 columns = ['feat1', 'feat2', 'feat3'])
    >>> synth_data = pd.DataFrame(get_synth(500),
                                  columns = ['feat1', 'feat2', 'feat3'])

    The first feature appears to be more similar than the second across
    datasets.

    >>> kullback_leibler(real_data, synth_data, 'feat1', bins = 20)
    0.03389133708660097 # random
    >>> kullback_leibler(real_data, synth_data, 'feat2', bins = 20)
    0.58739109417064730 # random

    """
    if bins is None:
        values = set(np.concatenate(
            [np.unique(real[feature]), np.unique(synth[feature])]))
        real_bincounts = [(real[feature].values == val).sum()
                          for val in values]
        synth_bincounts = [(synth[feature].values == val).sum()
                           for val in values]
    else:
        # assign bin edges based on joined data
        bin_edges = np.histogram_bin_edges(np.concatenate((real[feature],
                                                           synth[feature])),
                                           bins=bins)
        # count the number of points in each bin seperately
        real_bincounts = np.histogram(real[feature], bin_edges)[0]
        synth_bincounts = np.histogram(synth[feature], bin_edges)[0]
    return entropy(real_bincounts, synth_bincounts, **kwargs)


def jensen_shannon_divergence(real, synth, feature, bins='auto', **kwargs):
    """ Divergence: Jensen-Shannon

    The Jensen-Shannon divergence describes the difference between the `real`
    and `synth` distributions of the `feature` in terms of entropy. It is the
    square of Jensen-Shannon distance.

    Parameters
    ----------
    real : pandas dataframe
        Dataframe containing the real data.
    synth : pandas dataframe
        Dataframe containing the synthetic data.
    feature : str
        String indicating the feature of the datasets to compare. This must be
        a continuous variable.
    bins : int or str
        Number of bins to use when discretising data, if `None` data will be
        treated as categorical. If string chosen method will be used to
        calculate optimal bin width, 'auto' is the default. See
        numpy.histogram_bin_edges for further options.
    **kwargs : dict
        Keyword arguments.

    Returns
    -------
    divergence : double
        The computed divergence between the distributions.

    See Also
    --------
    distance.jensen_shannon_distance
    scipy.spatial.distance.jensenshannon

    Notes
    -----
    This is a wrapper function that just squares the result of
    `distance.jensen_shannon_distance`, which in turn wraps
    `scipy.spatial.distance.jensenshannon`. Since this function expects
    probability vectors the data is first discretised into evenly-spaced bins.

    We can think of the Jensen-Shannon divergence as the amount of information,
    or entropy, encoded in the difference between the `real` and `synth`
    distributions of the `feature`.

    The distance is zero if the distributions are identical, and is bounded
    above by one if they are nothing alike. This method is therefore good
    for comparing multiple synthetic datasets, or features within a dataset,
    to see which is closest to the real. However, as this is not a test,
    there is no threshold distance below which we can claim the distributions
    are statistically the same.

    An optimal 'bins' value has not been suggested.

    Examples
    --------
    >>> import pandas as pd
    >>> real_data = pd.DataFrame(get_real(500),
                                 columns = ['feat1', 'feat2', 'feat3'])
    >>> synth_data = pd.DataFrame(get_synth(500),
                                  columns = ['feat1', 'feat2', 'feat3'])

    The first feature appears to be more similar than the second across
    datasets.

    >>> jensen_shannon_divergence(real_data, synth_data, 'feat1', bins = 20)
    0.11006632967333475 # random
    >>> jensen_shannon_divergence(real_data, synth_data, 'feat2', bins = 20)
    0.43556476029981644 # random

    """
    return jensen_shannon_distance(real, synth, feature, bins, **kwargs) ** 2


def mann_whitney(real, synth, feature, **kwargs):
    """ Hypothesis Test: Mann-Whitney

    The Mann-Whitney test compares the distributions of data by examining how
    well-mixed they are when pooled. This is acheived by ranking the pooled
    data. A low p-value suggests the data are not similar.

    Parameters
    ----------
    real : pandas dataframe
        Dataframe containing the real data.
    synth : pandas dataframe
        Dataframe containing the synthetic data.
    feature : str
        String indicating the feature of the datasets to compare. This must be
        a continuous variable.
    **kwargs : dict
        Keyword arguments.

    Returns
    -------
    statistic : float
        The Mann-Whitney U statistic, in particular U for `synth`.
    pvalue : float
        Two-sided p-value assuming an asymptotic normal distribution.

    See Also
    --------
    scipy.stats.mannwhitneyu

    Notes
    -----
    This is a wrapper function for `scipy.stats.mannwhitneyu`.

    The null hypothesis for this test is that for randomly selected real and
    synthetic values, the probability that the real value is greater than the
    synthetic is the same as the probability that the synthetic value is
    greater than the real.

    We reject this hypothesis if the p-value is suitably small. This would in
    turn suggest that the synthetic and real data are not similarly
    distributed.

    Examples
    --------
    >>> import pandas as pd
    >>> real_data = pd.DataFrame(get_real(500),
    ...                          columns = ['feat1', 'feat2', 'feat3'])
    >>> synth_data = pd.DataFrame(get_synth(500),
    ...                            columns = ['feat1', 'feat2', 'feat3'])

    If we were to choose our p-value threshold as 0.05, we would reach the
    conclusion that the distributions of the first feature are similar but
    the distributions of the second feature are not.

    >>> mann_whitney(real_data, synth_data, 'feat1', bins = 20)
    MannwhitneyuResult(statistic=126910.0, pvalue=0.6758436855431454) # random
    >>> mann_whitney(real_data, synth_data, 'feat2', bins = 20)
    MannwhitneyuResult(statistic=134107.0, pvalue=0.04613704446362845) # random
    """
    return mannwhitneyu(real[feature], synth[feature],
                        alternative='two-sided', **kwargs)


def wilcoxon(real, synth, feature, **kwargs):
    """ Hypothesis Test: Wilcoxon

    The Wilcoxon test compares the distributions of paired data. It does this
    by ranking the differences.

    Parameters
    ----------
    real : pandas dataframe
        Dataframe containing the real data.
    synth : pandas dataframe
        Dataframe containing the synthetic data.
    feature : str
        String indicating the feature of the datasets to compare. This must be
        a continuous variable.
    **kwargs : dict
        Keyword arguments.

    Returns
    -------
    statistic : float
        The sum of the ranks of the differences above or below zero, whichever
        is greater.
    pvalue : float
        Two-sided p-value.

    See Also
    --------
    scipy.stats.wilcoxon

    Notes
    -----
    This is a wrapper function for `scipy.stats.wilcoxon`.

    The null hypothesis for this test is that the median of the paired
    differences is zero. The alternative hypothesis is that it is different
    from zero. This would suggest that the synthetic and real data are not
    similarly distributed.

    This test only makes sense when the synthetic and real data are paired.
    That is, each synthetic datapoint is matched to a real one. In which case,
    it is required that data are ordered to reflect this.

    Examples
    --------
    >>> import pandas as pd
    >>> real_data = pd.DataFrame(get_real(500),
    ...                          columns = ['feat1', 'feat2', 'feat3'])
    >>> synth_data = pd.DataFrame(get_synth(500),
    ...                            columns = ['feat1', 'feat2', 'feat3'])

    If we were to choose our p-value threshold as 0.05, we would reach the
    conclusion that the distributions of the first feature are similar but of
    the second feature are not.

    >>> wilcoxon(real_data, synth_data, 'feat1', bins = 20)
    WilcoxonResult(statistic=58917.0, pvalue=0.25131501183065175) # random
    >>> wilcoxon(real_data, synth_data, 'feat2', bins = 20)
    WilcoxonResult(statistic=54474.0, pvalue=0.011678503879013464) # random
    """
    return scipy_wc(real[feature], y=synth[feature], **kwargs)


def kruskal_wallis(real, synth, feature, **kwargs):
    """ Hypothesis Test: Kruskal-Wallis

    The Kruskal-Wallis test compares the distributions of data by examining how
    well-mixed they are when pooled. This is acheived by ranking the pooled
    data. A low p-value suggests the data are not similar.

    Parameters
    ----------
    real : pandas dataframe
        Dataframe containing the real data.
    synth : pandas dataframe
        Dataframe containing the synthetic data.
    feature : str
        String indicating the feature of the datasets to compare. This must be
        a continuous variable.
    **kwargs : dict
        Keyword arguments.

    Returns
    -------
    statistic : float
        The Kruskal-Wallis H statistic.
    pvalue : float
        The p-value for the test.

    See Also
    --------
    scipy.stats.kruskal

    Notes
    -----
    This is a wrapper function for `scipy.stats.kruskal`.

    The null hypothesis for this test is that the medians of the distributions
    are equal. The alternative hypothesis is then that they are different. This
    would suggest that the synthetic and real data are not similarly
    distributed.

    We notice however that failure to reject the null hypothesis only suggests
    that the medians could be equal and says nothing else about how the data
    are distributed.

    Examples
    --------
    >>> import pandas as pd
    >>> real_data = pd.DataFrame(get_real(500),
    ...                          columns = ['feat1', 'feat2', 'feat3'])
    >>> synth_data = pd.DataFrame(get_synth(500),
    ...                            columns = ['feat1', 'feat2', 'feat3'])

    If we were to choose our p-value threshold as 0.05, we would reach the
    conclusion that the distributions of the first feature are similar but the
    distributions of the second feature are not.

    >>> kruskal_wallis(real_data, synth_data, 'feat1', bins = 20)
    KruskalResult(statistic=1.4447530549450676, pvalue=0.22937173881858086)
    # random
    >>> kruskal_wallis(real_data, synth_data, 'feat2', bins = 20)
    KruskalResult(statistic=5.1566145854149, pvalue=0.023157995217201643)
    # random
    """
    return kruskal(real[feature], synth[feature], **kwargs)


if __name__ == '__main__':
    pass



from sklearn.metrics import mean_squared_error
import numpy as np
import pandas as pd
from numpy import corrcoef
from scipy.stats import chi2_contingency
from itertools import combinations, product
import warnings


def correlation_MSD(real, synth, feats=None):
    """correlation mean-squared-difference

    This metric calculates the mean squared difference between the correlation
    matrices for the real and synthetic datasets. This gives an indication of
    how well the synthetic data has retained inter-variable relationships.

    Parameters
    ----------
    real : pandas dataframe
        Dataframe containing the real data.
    synth : pandas dataframe
        Dataframe containing the synthetic data.
    feats: str or list of str, optional
        Numeric features to use. Non-numeric features will be filtered out. By
        default all numeric features are used.

    Returns
    -------
    MSD_c : float
        Correlation mean squared difference.

    See Also
    --------
    numpy.corrcoef

    Notes
    -----
    Correlations only make sense for numerical data, so this function first
    filters the columns by datatype.

    The smaller the number this function returns, the better the synthetic
    data captures the correlation between variables in the real data. This
    method is therefore good for comparing multiple synthetic datasets to see
    which is closest to the real. However, as this is not a test, there is no
    threshold value below which we can claim the datasets are statistically
    the same.
    """
    if isinstance(feats, pd.Index):
        feats = feats
    elif isinstance(feats, str):
        feats = [feats]
    else:
        feats = feats or real.columns.to_list()
    # select only numeric variables
    real_numeric = real[feats].select_dtypes(np.number)
    synth_numeric = synth[feats].select_dtypes(np.number)
    # calculate correlation matrices
    real_corr = corrcoef(real_numeric, rowvar=False)
    synth_corr = corrcoef(synth_numeric, rowvar=False)
    # return mean squared difference
    return mean_squared_error(real_corr, synth_corr)


def cramers_v(var1, var2):
    """Cramer's V

    Measures the association between two nominal categorical variables.

    Parameters
    ----------
    var1 : pandas.Series
        Series object containing the values for one of the variables to be used
        in the comparison.
    var2 : pandas.Series
        Series object containing the values for the other variable to be used
        in the comparison.

    Returns
    -------
    Cramers_V : float
        The amount of association between the two variables.

    Notes
    -----
    Wikipedia suggests that this formulation of Cramer's V tends to
    overestimate the strength of an association and poses a corrected version.
    However, since we are only concerned with how associations compare and not
    what the actual values are, we continue to use this simpler version.
    """
    confusion_matrix = np.array(pd.crosstab(var1, var2))
    # Chi-squared test statistic, sample size, and minimum of rows and columns
    X2 = chi2_contingency(confusion_matrix, correction=False)[0]
    n = np.sum(confusion_matrix)
    minDim = min(confusion_matrix.shape)-1
    # Calculate Cramer's V
    return(np.sqrt((X2/n) / minDim))


def cramers_v_MSE(real, synth, feats=None):
    """Cramer's V Mean Squared Error

    This metric calculates the difference in association between categorical
    features in the real and synthetic datasets.

    Parameters
    ----------
    real : pandas dataframe
        Dataframe containing the real data.
    synth : pandas dataframe
        Dataframe containing the synthetic data.
    feats: str or list of str, optional
        Feature(s) in `real` and `synth` to include in comparison. By default
        all object and categorical columns are selected.

    Returns
    -------
    cramers_v_MSE : float
        Mean squared error between `real` and `synth` in Cramer's V scores
        across feature pairs.

    Notes
    -----
    This metric is only valid for categorical features so a warning is returned
    if any of the selected features appear to be numeric. If no features are
    selected, only the `category` and `object` types are used.
    """
    # check features are categorical if supplied
    if isinstance(feats, str):
        feats = [feats]
    if feats is None:
        feats = real.select_dtypes(include=['object', 'category']).columns
    else:
        non_cat = real[feats].select_dtypes(
            exclude=['object', 'category']).columns
        if len(non_cat) > 0:
            warnings.warn('Selected features include numeric types:'
                          f'{non_cat} If these should not be included, rerun '
                          'specifying different features. Otherwise, they will'
                          ' be assumed to be encoded categories.')
    # find all possible feature combinations
    feat_combinations = list(combinations(feats, 2))
    real_cramers_v = []
    synth_cramers_v = []
    # loop through feature combinations
    for feat1, feat2 in feat_combinations:
        real_cramers_v.append(cramers_v(real[feat1], real[feat2]))
        synth_cramers_v.append(cramers_v(synth[feat1], synth[feat2]))
    return mean_squared_error(real_cramers_v, synth_cramers_v)


def correlation_ratio(categorical, continuous):
    """Correlation ratio

    Calculates the correlation ratio for categorical-continuous association.
    Describes the possibility of deducing the corresponding category for a
    given continuous value.

    Returns a value in the range [0,1] where 0 means a category can not be
    determined given a continuous measurement and 1 means it can with absolute
    certainty.

    Parameters
    ----------
    categorical : Pandas Series
        A sequence of categorical measurements
    continuous : Pandas Series
        A sequence of continuous measurements

    Returns
    -------
    float in the range [0,1]

    Notes
    -----
    See https://en.wikipedia.org/wiki/Correlation_ratio for more details.
    """
    categories = pd.unique(categorical)
    category_means = np.zeros(len(categories))
    category_counts = np.zeros(len(categories))
    for i, cat in enumerate(categories):
        cts_in_cat = continuous[categorical == cat]
        category_means[i] = np.mean(cts_in_cat)
        category_counts[i] = len(cts_in_cat)
    total_mean = np.mean(continuous)
    numerator = np.sum(category_counts*((category_means-total_mean)**2))
    denominator = np.sum((continuous-total_mean)**2)
    return np.sqrt(numerator/denominator)


def correlation_ratio_MSE(real, synth, categorical_feats='auto',
                          numerical_feats=None):
    """Correlation Ratio Mean Squared Error

    This metric calculates the difference in association between categorical
    and continuous feature pairings in the real and synthetic datasets.

    Parameters
    ----------
    real : pandas dataframe
        Dataframe containing the real data.
    synth : pandas dataframe
        Dataframe containing the synthetic data.
    categorical_feats: str or list of str, optional
        Categorical feature(s) in `real` and `synth` to include in comparison.
        By default all object and categorical columns are selected.
    numerical_feats: str or list of str, optional
        Numerical feature(s) in `real` and `synth` to include in comparison.
        By default all columns not in `categorical_feats` are selected.

    Returns
    -------
    corr_ratio_MSE : float
        Mean squared error between `real` and `synth` in correlation ratio
        scores across valid feature pairs.

    Notes
    -----
    If no categorical features are selected, columns of type `category` or
    `object` are used. If no numerical features are selected, all columns that
    are not listed as categorical features are used.
    """
    if categorical_feats == 'auto':
        categorical_feats = real.select_dtypes(
            include=['object', 'category']).columns
    if numerical_feats is None:
        numerical_feats = list(set(real.columns)-set(categorical_feats))
    real_corr_ratio = []
    synth_corr_ratio = []
    for cat_feat, num_feat in list(product(categorical_feats,
                                           numerical_feats)):
        real_corr_ratio.append(correlation_ratio(
            real[cat_feat], real[num_feat]))
        synth_corr_ratio.append(correlation_ratio(
            synth[cat_feat], synth[num_feat]))
    return mean_squared_error(real_corr_ratio, synth_corr_ratio)


if __name__ == '__main__':
    pass



import numpy as np
import pandas as pd
from sklearn.linear_model import LogisticRegression
from sklearn.tree import DecisionTreeClassifier
from sklearn.preprocessing import PolynomialFeatures
import random
import warnings
from collections import namedtuple
from ..utils import df_combine


def propensity_MSE(real, synth, method, **kwargs):
    """Propensity mean-squared error

    We think of the propensity of an example as the estimated probability it is
    classified as being synthetic. For a good synthetic dataset we would expect
    this to be the same as the proportion of synthetic examples used in
    training. The mean squared error from this proportion is hence a suitable
    utility metric.

    Parameters
    ----------
    real : pandas dataframe
        Dataframe containing the real data.
    synth : pandas dataframe
        Dataframe containing the synthetic data.
    method : str, ['CART', 'LogisticRegression']
        Classification method to use.
    **kwargs : dict
        Keyword arguments passed to classification function.

    Returns
    -------
    MSE_p : float
        Propensity score mean squared error.

    See Also
    --------
    sklearn.linear_model.LogisticRegression
    sklearn.tree.DecisionTreeClassifier

    Notes
    -----
    Propensity scores represent probabilities of group membership. By modelling
    whether an example is synthetic or not, we can use propensity scores as a
    measure of utility.

    This returns zero if the distributions are identical, and is bounded
    above by 1-c if they are nothing alike, where c is the proportion of
    synthetic data. This method is therefore good for comparing multiple
    synthetic datasets. However, as this is not a test, there is no threshold
    distance below which we can claim the distributions are statistically the
    same.

    This function assumes that some preprocessing has been carried out so that
    the data is ready to be passed to the classification function. Encoding
    of categorical data is performed, but, for example, scaling is not. Without
    this erroneous results may be returned. The logistic regression can fail to
    converge if many variables are considered. Anecdotally, this doesn't seem
    to drastically impact the propensity scores, although this should be
    investigated formally. `**kwargs` are passed to the classification model so
    it can be tuned.

    Using a CART model as a classifier is recommended in the literature however
    we also support the use of logistic regression.

    https://rss.onlinelibrary.wiley.com/doi/pdf/10.1111/rssa.12358
    """
    if method not in ['CART', 'LogisticRegression']:
        raise ValueError(
            "method must be either 'CART' or 'LogisticRegression'")
    # combine data
    combined = df_combine(real, synth, source_val_real=0, source_val_synth=1)
    # remove source column
    synth_bool = combined.pop('source')
    # encode categorical variables
    combined_encoded = pd.get_dummies(combined, drop_first=True)
    if method == 'LogisticRegression':
        # add interactions
        combined_encoded = PolynomialFeatures(2, interaction_only=True,
                                              include_bias=False) \
            .fit_transform(combined_encoded)

        model = LogisticRegression(**kwargs)
    if method == 'CART':
        model = DecisionTreeClassifier(**kwargs)
    model.fit(combined_encoded, synth_bool)

    # calculate propensity for each example
    props = model.predict_proba(combined_encoded)[:, 1]

    ideal_prop = len(synth)/len(combined_encoded)

    props_square_error = np.square(props - ideal_prop)

    MSE_p = sum(props_square_error)/len(props_square_error)

    return MSE_p


def expected_p_MSE(real, synth):
    """ Expected propensity mean-squared error

    This is the expected propensity mean-squared error under the null case
    that the `real` and `synth` datasets are distributionally similar.

    Parameters
    ----------
    real : pandas dataframe
        Dataframe containing the real data.
    synth : pandas dataframe
        Dataframe containing the synthetic data.

    Returns
    -------
    MSE_p : float
        Expected propensity score mean square-squared error.

    Notes
    -----
    This expectation is used to standardise `propensity_MSE` to an
    interpretable scale.

    It has been shown that the null case is distributed as a multiple of a
    :math:`\\chi^2` distribution with :math:`k-1` degrees of freedom, which has
    expectation:

    .. math::

        (k-1)(1-c)^2c/N

    where :math:`k` is the number of predictors in the model, :math:`c` is the
    proportion of synthetic data, and :math:`N` is the total number of examples
    Further explanation and derivation of this formulation can be found here:
    https://rss.onlinelibrary.wiley.com/doi/pdf/10.1111/rssa.12358 Appendix A1.

    """
    # =(k−1)(1−c)^2*c/N
    num_vars = real.shape[1]
    num_vars_and_interactions = num_vars*(num_vars+1)/2
    total_num_examples = real.shape[0]+synth.shape[0]
    prop_synth = synth.shape[0]/total_num_examples
    return ((num_vars_and_interactions)*(1.0-prop_synth)
            ** 2*prop_synth/total_num_examples)


def stdev_p_MSE(real, synth):
    """ Standard deviation propensity mean-squared error

    This is the standard deviation of the propensity mean-squared error under
    the null case that the `real` and `synth` datasets are distributionally
    similar.

    Parameters
    ----------
    real : pandas dataframe
        Dataframe containing the real data.
    synth : pandas dataframe
        Dataframe containing the synthetic data.

    Returns
    -------
    st_dev : float
        Expected propensity score mean square-squared error standard deviation.

    Notes
    -----
    This standard deviation is used to standardise `propensity_MSE` to an
    interpretable scale.

    It has been shown that the null case is distributed as a multiple of a
    :math:`\\chi^2` distribution with :math:`k-1` degrees of freedom, which has
    standard deviation:

    .. math::

        \\sqrt{2(k-1)}(1-c)^2c/N

    where :math:`k` is the number of predictors in the model, :math:`c` is the
    proportion of synthetic data, and :math:`N` is the total number of examples
    Further explanation and derivation of this formulation can be found here:
    https://rss.onlinelibrary.wiley.com/doi/pdf/10.1111/rssa.12358 Appendix A1.
    """
    num_vars = real.shape[1]
    num_vars_and_interactions = num_vars*(num_vars+1)/2
    total_num_examples = real.shape[0]+synth.shape[0]
    prop_synth = synth.shape[0]/total_num_examples
    return ((2*(num_vars_and_interactions))
            ** 0.5*(1.0-prop_synth)**2*prop_synth/total_num_examples)


def perm_expected_sd_p_MSE(real, synth, num_perms=20, **kwargs):
    """Permutation Expected and Standard Deviation Propensity Mean Squared Error

    Repeat pMSE calculation several times but while randomly permutating the
    boolean indicator column. This should approximate propensity mean squared
    error results for 'properly' synthesised datasets when it is not possible
    to calculate these directly, in particular when CART is used.

    Parameters
    ----------
    real : pandas dataframe
        Dataframe containing the real data.
    synth : pandas dataframe
        Dataframe containing the synthetic data.
    num_perms : int
        The number of times to repeat the process.
    **kwargs : dict
        Keyword arguments passed to DecisionTreeClassifier function.

    Returns
    -------
    mean : float
        Mean of the propensity_MSE scores over all the repititions.
    sd : float
        Standard deviation of the propensity_MSE scores over all the
        repititions.

    Notes
    -----
    This function is only intended to be used within `propensity_metrics()`.

    This function is stochastic so will return a differnet result every time.
    """
    perm_MSEs = np.zeros(num_perms)
    for i in range(num_perms):
        combined = df_combine(
            real, synth, source_val_real=0, source_val_synth=1)
        # remove source column
        synth_bool = combined.pop('source').tolist()
        random.shuffle(synth_bool)
        # encode categorical variables
        combined_encoded = pd.get_dummies(combined, drop_first=True)
        model = DecisionTreeClassifier(**kwargs)
        model.fit(combined_encoded, synth_bool)
        # calculate propensity for each example
        props = model.predict_proba(combined_encoded)[:, 1]

        ideal_prop = len(synth)/len(combined_encoded)

        props_square_error = np.square(props - ideal_prop)

        MSE_p = sum(props_square_error)/len(props_square_error)

        perm_MSEs[i] = MSE_p
    return np.mean(perm_MSEs), np.std(perm_MSEs)


def p_MSE_ratio(real, synth, method='CART', feats=None, **kwargs):
    """ Propensity mean-squared error ratio

    This is the ratio of observed propensity mean-squared error, to that
    expected under the null case that the synthetic and real datasets are
    similarly distributed.

    Parameters
    ----------
    real : pandas dataframe
        Dataframe containing the real data.
    synth : pandas dataframe
        Dataframe containing the synthetic data.
    method : str, ['CART', 'LogisticRegression']
        Classification method to use.
    feats : list of str, optional.
        List of features in the dataset that will be used in propensity
        calculations. By default all features will be used.
    **kwargs : dict
        Keyword arguments passed to classifier function.

    Returns
    -------
    ratio : float
        Ratio of the propensity score mean square-squared error.

    Notes
    -----
    This standardisation transforms the scale for the MSE into one that makes
    more sense for synthetic data. Before, the MSE score gave better utility
    as the value got closer to zero, which is only attainable when the datasets
    are identical. However, when generating synthetic data we do not want to
    produce identical entries, but to acheive distributional similarity between
    the distribution of the observed data and the model used to generate the
    synthetic.

    This ratio tends towards one when the datasets are similar and increases
    otherwise.
    """
    warnings.simplefilter("always", category=DeprecationWarning)
    warnings.warn(
        "p_MSE_ratio is now contained within `propensity_metrics`.",
        DeprecationWarning)
    if isinstance(feats, pd.Index):
        feats = feats
    elif isinstance(feats, str):
        feats = [feats]
    else:
        feats = feats or real.columns.to_list()
    if method == 'Logistic Regression':
        exp_p_MSE = expected_p_MSE(real[feats], synth[feats])
    if method == 'CART':
        exp_p_MSE, _ = perm_expected_sd_p_MSE(
            real[feats], synth[feats], num_perms=2, **kwargs)
    obs_p_MSE = propensity_MSE(
        real[feats], synth[feats], method=method, **kwargs)
    return obs_p_MSE/exp_p_MSE


def standardised_p_MSE(real, synth, method='CART', feats=None, **kwargs):
    """ Standardised propensity mean-squared error

    This is the difference between the observed propensity mean-squared error
    and that expected under the null case that the synthetic and real datasets
    are similarly distributed, scaled by the standard deviation.

    Parameters
    ----------
    real : pandas dataframe
        Dataframe containing the real data.
    synth : pandas dataframe
        Dataframe containing the synthetic data.
    method : str, ['CART', 'LogisticRegression']
        Classification method to use.
    feats : list of str, optional.
        List of features in the dataset that will be used in propensity
        calculations. By default all features will be used.
    **kwargs : dict
        Keyword arguments passed to LogisticRegression function.

    Returns
    -------
    ratio : float
        Standardised propensity score mean square-squared error.

    Notes
    -----
    This standardisation transforms the scale for the MSE into one that makes
    more sense for synthetic data. Before, the MSE score gave better utility
    as the value got closer to zero, which is only attainable when the datasets
    are identical. However, when generating synthetic data we do not want to
    produce identical entries, but to acheive distributional similarity between
    the distribution of the observed data and the model used to generate the
    synthetic.

    This metric tends towards zero when the datasets are similar and increases
    otherwise.
    """
    warnings.simplefilter("always", category=DeprecationWarning)
    warnings.warn(
        "standardised_p_MSE is now contained within `propensity_metrics`.",
        DeprecationWarning)
    if isinstance(feats, pd.Index):
        feats = feats
    elif isinstance(feats, str):
        feats = [feats]
    else:
        feats = feats or real.columns.to_list()

    if method == 'LogisticRegression':
        exp_p_MSE = expected_p_MSE(real[feats], synth[feats])
        std_p_MSE = stdev_p_MSE(real[feats], synth[feats])

    if method == 'CART':
        exp_p_MSE, std_p_MSE = perm_expected_sd_p_MSE(
            real[feats], synth[feats], **kwargs)

    obs_p_MSE = propensity_MSE(
        real[feats], synth[feats], method=method, **kwargs)
    return (obs_p_MSE - exp_p_MSE)/std_p_MSE


def propensity_metrics(real, synth, method='CART', feats=None, num_perms=20,
                       **kwargs):
    """ Propensity metrics

    This function calculates three flavours of propensity mean-squared error,
    all of which quantify utility by measuring how well a classifier can be
    trained to distinguish `real` and `synth`. It returns the raw, observed
    value together with this value standardised by the expected value if the
    dataset was well synthesised, and also as a ratio of the expected value.

    Parameters
    ----------
    real : pandas dataframe
        Dataframe containing the real data.
    synth : pandas dataframe
        Dataframe containing the synthetic data.
    method : str, ['CART', 'LogisticRegression']
        Classification method to use.
    feats : list of str, optional.
        List of features in the dataset that will be used in propensity
        calculations. By default all features will be used.
    num_perms : int
        If CART method is used, this is the number of times to repeat process
        when calculating the expected mean-squared error.
    **kwargs : dict
        Keyword arguments passed to LogisticRegression function.

    Returns
    -------
    observed_p_MSE : float
        Observed propensity score mean squared error.
    standardised_p_MSE : float
        Standardised propensity score mean square-squared error.
    ratio_p_MSE : float
        Ratio of the propensity score mean square-squared error to the
        expected value.


    Notes
    -----
    The standardisation and ratio operations transform the scales for the MSE
    into ones that make more sense for synthetic data. For the observed
    propensity the MSE score gives better utility as the value gets closer to
    zero, which is only attainable when the datasets are identical. However,
    when generating synthetic data we do not want to produce identical entries,
    but to acheive distributional similarity between the distribution of the
    observed data and the model used to generate the synthetic.

    The standardised score tends towards zero when the datasets are similar
    and increases otherwise.

    The ratio tends towards one when the datasets are similar and increases
    otherwise.
    """
    if isinstance(feats, pd.Index):
        feats = feats
    elif isinstance(feats, str):
        feats = [feats]
    else:
        feats = feats or real.columns.to_list()

    if method == 'LogisticRegression':
        exp_p_MSE = expected_p_MSE(real[feats], synth[feats])
        std_p_MSE = stdev_p_MSE(real[feats], synth[feats])

    if method == 'CART':
        exp_p_MSE, std_p_MSE = perm_expected_sd_p_MSE(
            real[feats], synth[feats], num_perms=num_perms, **kwargs)

    obs_p_MSE = propensity_MSE(
        real[feats], synth[feats], method=method, **kwargs)
    standardised_p_MSE = (obs_p_MSE - exp_p_MSE)/std_p_MSE
    ratio_p_MSE = obs_p_MSE/exp_p_MSE

    PropensityResult = namedtuple('PropensityResult', ('observed_p_MSE',
                                                       'standardised_p_MSE',
                                                       'ratio_p_MSE'))

    return PropensityResult(obs_p_MSE, standardised_p_MSE, ratio_p_MSE)


if __name__ == '__main__':
    pass



from collections import namedtuple
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler, OneHotEncoder
from sklearn.pipeline import Pipeline
from sklearn.compose import ColumnTransformer
from sklearn.metrics import precision_score, recall_score, f1_score


def classification_comparison(real, synth, key, target, sklearn_classifier,
                              random_state=None, test_prop=0.2, **kwargs):
    """Classification utility metric

    This metric fits two classification models to `real` and `synth`
    respectively, and then tests them both against withheld `real` data. We
    obtain utility scores by subtracting the precision, recall and f1 scores
    of the predictions obtained by the synth model from those obtained by the
    real model.

    Parameters
    ----------
    real : pandas dataframe
        Dataframe containing the real data.
    synth : pandas dataframe
        Dataframe containing the synthetic data.
    key : list of strings
        list of column names to use as the input in the classification.
    target : str
        column to use as target in the classification.
    sklearn_classifier : scikit-learn estimator
        classifier with fit and predict methods.
    random_state : int, RandomState instance or None, default=42
        Controls the shuffling steps during the train-test split and the
        classification algorithm itself. Pass an int for reproducible output
        across multiple function calls.
    test_prop : float or int, default=0.2
        If float, should be between 0.0 and 1.0 and represent the proportion
        of the dataset to include in the test split. If int, represents the
        absolute number of test samples.

    Returns
    -------
    ClassificationResult : namedtuple
        precision_difference : float
            precision of model trained on real data subtracted by precision of
            model trained on synthetic data.
        recall_difference : float
            recall of model trained on real data subtracted by recall of
            model trained on synthetic data.
        f1_difference : float
            f1 score of model trained on real data subtracted by f1 score of
            model trained on synthetic data.

    Notes
    -----
    Some preprocessing is carried out before the models are trained. Numeric
    features are scaled and categorical features are one-hot-encoded.

    A score of zero tells us the synthetic data is just as good as the real at
    training classifier models. Increases in these scores indicate poorer
    utility.
    """
    real_X = real[key]
    real_y = real[target]

    synth_X = synth[key]
    synth_y = synth[target]

    # split data
    real_X_train, real_X_test, real_y_train, real_y_test = train_test_split(
        real_X, real_y, test_size=test_prop, random_state=random_state)
    synth_X_train, _, synth_y_train, _ = train_test_split(
        synth_X, synth_y, test_size=test_prop, random_state=random_state)

    # preprocessing
    # scale numeric
    numeric_transformer = Pipeline(steps=[('scaler', StandardScaler())])
    # OHE categorical
    categor_transformer = Pipeline(steps=[('encoder', OneHotEncoder())])

    num_feats = real[key].select_dtypes(include='number').columns
    cat_feats = real[key].select_dtypes(exclude='number').columns

    preprocessor = ColumnTransformer(
        transformers=[
            ('numeric', numeric_transformer, num_feats),
            ('categor', categor_transformer, cat_feats)
        ]
    )
    # train real model
    real_pipeline = Pipeline(steps=[
        ('preprocessor', preprocessor),
        ('classifier', sklearn_classifier(random_state=random_state, **kwargs))
    ])

    real_pipeline.fit(real_X_train, real_y_train.values.ravel())
    # test real on real
    y_real_predicts_real = real_pipeline.predict(real_X_test)
    # train synth model
    synth_pipeline = Pipeline(steps=[
        ('preprocessor', preprocessor),
        ('classifier', sklearn_classifier(random_state=random_state, **kwargs))
    ])

    synth_pipeline.fit(synth_X_train, synth_y_train.values.ravel())
    # test synth on real
    y_synth_predicts_real = synth_pipeline.predict(real_X_test)

    # compare results
    real_precision = precision_score(
        real_y_test, y_real_predicts_real, average='macro')
    real_recall = recall_score(
        real_y_test, y_real_predicts_real, average='macro')
    real_f1 = f1_score(real_y_test, y_real_predicts_real, average='macro')

    synth_precision = precision_score(
        real_y_test, y_synth_predicts_real, average='macro')
    synth_recall = recall_score(
        real_y_test, y_synth_predicts_real, average='macro')
    synth_f1 = f1_score(real_y_test, y_synth_predicts_real, average='macro')

    diff_precision = real_precision - synth_precision
    diff_recall = real_recall - synth_recall
    diff_f1 = real_f1 - synth_f1

    ClassificationResult = namedtuple('ClassificationResult',
                                      ('precision_difference',
                                       'recall_difference',
                                       'f1_difference'))

    return ClassificationResult(diff_precision, diff_recall, diff_f1)


if __name__ == '__main__':
    pass



"""
clustering metric
"""
from sklearn.cluster import KMeans
from kmodes.kprototypes import KPrototypes
import numpy as np
from ..utils import df_combine


def clustered_MSD(combined, synthetic_indicator, method, k, random_state=None):
    """clustered mean-squared difference

    This metric performs a cluster analysis on the data and then considers the
    proportion of synthetic data in each cluster. It measures the difference
    between this and the overall proportion, squares, and then takes the mean
    across clusters.

    Parameters
    ----------
    combined : 2d numpy array
        Array containing both the real and synthetic data. Rows are examples
        and columns are variables.
    synthetic_indicator : 1d numpy array
        Corresponding boolean array indicating which of the rows in `combined`
        are synthetic (1), and which are real (0).
    method : {'kmeans','kprototypes'}, default='kmeans'
        Clustering method to use. See sklearn.cluster.KMeans and
        kmodes.kprototypes.KPrototypes for details.
    k : int
        Integer indicating how many clusters to form during the analysis.
    random_state : int, RandomState instance or None, default: None
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    Returns
    -------
    MSD_p : float
        Mean-squared difference between the within-cluster proportions of
        synthetic data and the overall proportion of synthetic data.

    Notes
    -----
    There is no obvious criterion for selecting the number of clusters `k`. One
    approach when comparing two different synthetic datasets against eachother
    could be to try several values of `k` with the original data to examine the
    sensitivity of the choice of `k`, and then go with the one that is most
    often the best choice.

    Additionally, it is important to note that this metric says nothing about
    how the data are distributed within clusters.
    """
    if method == 'kmeans':
        combined_numeric = combined.select_dtypes(include='number')
        clusters = KMeans(n_clusters=k, random_state=random_state).fit(
            combined_numeric).labels_
    if method == 'kprototype':
        cat_cols = combined.select_dtypes(
            include=['object', 'category']).columns
        cat_cols_i = [n for n, col in enumerate(
            combined.columns) if col in cat_cols]
        clusters = KPrototypes(n_clusters=k, random_state=random_state).fit(
            combined, categorical=cat_cols_i).labels_
    cluster_proportion = []
    for ki in range(k):
        cluster_proportion.append(sum(synthetic_indicator[clusters == ki])
                                  / sum(clusters == ki))
    cluster_proportion = np.array(cluster_proportion)

    # calculate error from ideal proportion
    ideal_prop = sum(synthetic_indicator)/len(combined)

    prop_square_error = np.square(cluster_proportion - ideal_prop)

    MSE_p = sum(prop_square_error)/len(prop_square_error)
    # print(MSE_p)
    return MSE_p


def multi_clustered_MSD(real, synth, feats=None, method='kmeans', k_min=10,
                        k_max=40, random_state=None):
    """multiple clustered mean-squared difference

    This metric performs the `clustered_MSD` above multiple times.

    Parameters
    ----------
    real : pandas dataframe
        Dataframe containing the real data.
    synth : pandas dataframe
        Dataframe containing the synthetic data.
    feats: str or list of str, optional
        Numeric features to use. Non-numeric features will be filtered out. By
        default all numeric features are used.
    method : {'kmeans','kprototypes'}, default='kmeans'
        Clustering method to use. See sklearn.cluster.KMeans and
        kmodes.kprototypes.KPrototypes for details.
    k_min : int
        Minimum number of clusters to perform cluster analysis with.
    k_max : int
        Maximum number of clusters to perform cluster analysis with.
    random_state : int, RandomState instance or None, default: None
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    Returns
    -------
    cluster_MSDs : 1d numpy array
        array containing MSDs from cluster analyses with different number of
        clusters

    Notes
    -----
    Since there is no obvious way to select the number of clusters `k`, one
    approach when comparing two different synthetic datasets against eachother
    could be to try several values of `k` with the original data to examine the
    sensitivity of the choice of `k`, and then go with the one that is most
    often the best choice.

    Additionally, it is important to note that this metric says nothing about
    how the data are distributed within clusters.

    """
    # combine data
    combined = df_combine(real, synth, feats=feats,
                          source_val_real=0, source_val_synth=1)
    # remove source column
    synth_bool = combined.pop('source')

    cluster_MSDs = []

    for k in range(k_min, k_max):
        # print(k)
        cluster_MSDs.append(clustered_MSD(
            combined, synth_bool, method, k, random_state=random_state))

    return min(cluster_MSDs)


if __name__ == '__main__':
    pass



from itertools import product
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from pandas import crosstab, cut, DataFrame, Index
from pandas.core.dtypes.common import (is_categorical_dtype, is_numeric_dtype,
                                       is_object_dtype)
import seaborn as sns
from .utils import cat_encode, feature_density_diff
from .metrics.correlation import cramers_v

sns.set_theme()


def suggest_label_rotation(ax, axis='x', char_lim=5):
    """Advise if labels should be rotated.

    For the specifed axis the maximum characters for the
    labels is calculated. If this is above `char_lim` the function
    will return `True` to suggest that labels should be rotated

    Parameters
    ----------
    ax: matplotlib.axes.Axes
        Axes to analyse.
    axis: str, optional
        Default is 'x'.
    char_lim: int, optional
        Default is `5`.

    Returns
    -------
    bool
        True if labels should be rotated, otherwise False.

    """
    tick_func = f'get_{axis.lower()}ticklabels'
    labels = getattr(ax, tick_func)()
    max_chars = max([len(label.get_text()) for label in labels])
    return max_chars > char_lim


def plot_histograms(df, feats=None, groupby=None, figcols=2, figsize=None):
    """Plot feature distributions.

    Plot a histogram (or countplot for categorical data) for each feature.
    Where multiple features are provided a grid will be generated to store all
    plots.

    Optionally, a groupby feature can be specified to apply a grouping prior
    to calculating the distribution

    Parameters
    ----------
    df: pandas.DataFrame
    feats: str or list of str
        Feature(s) to plot. Must be column(s) in `df`.
    groupby: str, default=None
        Feature on which to group data. Default value of ``None`` means no
        grouping is used.
    figcols: int, default=2
        Number of columns to use in figure. Only used when `feats` contains
        more than one feature.
    figsize: tuple of float, optional
        Size of figure in inches (W,H). Defaults to
        ``rcParams["figure.figsize"](default: [6.4, 4.8])``.

    Returns
    -------
    matplotlib.figure.Figure

    """

    if isinstance(feats, Index):
        feats = feats
    elif isinstance(feats, str):
        feats = [feats]
    else:
        feats = feats or df.columns

    n_rows = int(np.ceil(len(feats) / figcols))
    fig, axes = plt.subplots(n_rows, figcols, figsize=figsize)

    for ft, ax in zip(feats, axes.ravel()):
        if is_categorical_dtype(df[ft]) or is_object_dtype(df[ft]):
            dist_func = sns.countplot
        elif is_numeric_dtype(df[ft]):
            dist_func = sns.histplot

        dist_func(data=df, x=ft, ax=ax, hue=groupby)

    # Turn off axes with no data
    for ax in axes.ravel():
        if not ax.has_data():
            ax.set_visible(False)

    fig.tight_layout()

    return fig


def plot_joint(df, x, y, groupby=None, x_bins='auto', y_bins='auto',
               figsize=None):
    """Plot bivariate and univariate graphs.

    Convenience function that leverages seaborn. For more granular control
    refer to ``seaborn.JointGrid`` and ``seaborn.jointplot``.

    Parameters
    ----------
    df: pandas.DataFrame
        DataFrame containing the feature(s) to plot.
    x: str
        Feature to plot on x axis and and margin. Must be column in `df`.
    y: str
        Feature to plot on y axis and and margin. Must be column in `df`.
    groupby: str, default=None
        Feature on which to group data. Default value of ``None`` means no
        grouping is used.
    x_bins, y_bins: array_like or int or str, default='auto'
        If ``array_like`` must be sequence of bin edges. If ``int``, specifies
        the number of bins to use. If ``str``, specifies the method used to
        calculate the optimal bin width. Default=``auto``. For all possible
        values see ``numpy.histogram_bin_edges``.
    figsize: int, default=None
        Size of each side of the figure in inches (it will be square). Default
        value of ``None`` will use default seaborn setting.

    Returns
    -------
    seaborn.axisgrid.JointGrid

    """
    figsize = figsize or 6
    g = sns.JointGrid(height=figsize)

    # If a feature is categorical it must be used 'as_ordered' for plotting a
    # histogram
    def order(ft): return ft.cat.as_ordered() if hasattr(ft, 'cat') else ft
    sns.histplot(data=df, x=order(df[x]), y=order(df[y]), hue=groupby,
                 alpha=0.5, ax=g.ax_joint)

    # For margins can use countplot or hist depending on data type.
    # No legends are shown for these marginal plots.
    if is_categorical_dtype(df[x]) or is_object_dtype(df[x]):
        ax = sns.countplot(data=df, x=x, hue=groupby, ax=g.ax_marg_x)
        lg = ax.get_legend()
        if lg is not None:
            lg.remove()
    else:
        sns.histplot(data=df, x=x, hue=groupby, bins=x_bins, ax=g.ax_marg_x,
                     legend=False)

    if is_categorical_dtype(df[y]) or is_object_dtype(df[y]):
        ax = sns.countplot(data=df, y=y, hue=groupby, ax=g.ax_marg_y)
        lg = ax.get_legend()
        if lg is not None:
            lg.remove()
    else:
        sns.histplot(data=df, y=y, hue=groupby, bins=y_bins, ax=g.ax_marg_y,
                     legend=False)

    return g


def plot_histogram3d(df, x, y, x_bins='auto', y_bins='auto', figsize=None):
    """Plot 3D histogram of two features.

    This is similar to a 2D histogram plot with an extra axis added
    to display the count for each featurewise pair as 3D bars.

    Parameters
    ----------
    df: pandas.DataFrame
        DataFrame containing the feature(s) to plot.
    x: str
        Feature to plot on x axis and and margin. Must be column in `df`.
    y: str
        Feature to plot on y axis and and margin. Must be column in `df`.
    x_bins, y_bins: array_like or int or str, default='auto'
        If ``array_like`` must be sequence of bin edges. If ``int``, specifies
        the number of bins to use. If ``str``, specifies the method used to
        calculate the optimal bin width. Default= ``auto``. For all possible
        values see ``numpy.histogram_bin_edges``. If `x` or `y` is categorical
        the bins will be set to the cardinality of the feature.
    figsize: int, default=None
        Size of each side of the figure in inches (it will be square). Default
        value of ``None`` will use default plot settings.

    Returns
    -------
    matplotlib.figure.Figure
    """
    # Encode categorical data
    cat_feats = [ft_name for ft_name, ft_series in df.iteritems()
                 if is_categorical_dtype(ft_series)
                 or is_object_dtype(ft_series)]
    cat_labels = dict()

    if len(cat_feats) > 0:
        df, cat_labels = cat_encode(df, cat_feats, return_all=True)

    # Determine bins
    bins_xy = []
    for ft, ft_bins in zip([x, y], [x_bins, y_bins]):
        if ft in cat_feats:
            # If categorical set bins to cardinality
            bins_xy.append(np.histogram_bin_edges(df[ft], df[ft].nunique()))
        else:
            # otherwise compute bins as specified
            bins_xy.append(np.histogram_bin_edges(df[ft], ft_bins))

    # Create figure
    fig = plt.figure(figsize=figsize)
    ax = plt.axes(projection='3d')

    # Compute 2D histogram
    hist, xedges, yedges = np.histogram2d(df[x], df[y], bins=bins_xy)

    # Determine bar anchors and thickness
    xpos, ypos = np.meshgrid(xedges[:-1], yedges[:-1], indexing="ij")
    xpos = xpos.ravel()
    ypos = ypos.ravel()

    dx = (xedges[1] - xedges[0]) * 0.8
    dy = (yedges[1] - yedges[0]) * 0.8
    dz = hist.ravel()
    zpos = np.zeros_like(dz) + 0.1

    # Plot the 3D bars
    ax.bar3d(xpos, ypos, zpos, dx, dy, dz, zsort='average')

    # Set title and labels
    ax.set_xlabel(x, fontsize=16)
    ax.set_ylabel(y, fontsize=16)
    ax.set_zlabel('$count$', fontsize=16, rotation=1)
    ax.set_title('3D Histogram')

    # Set tick lables for categorical features
    if x in cat_labels.keys():
        ax.set_xticks(xedges[1:])
        ax.set_xticklabels(cat_labels[x])
    if y in cat_labels.keys():
        ax.set_yticks(yedges[1:])
        ax.set_yticklabels(cat_labels[y])

    return fig


def plot_correlation(*df, feats=None, method='pearson', plot_diff=False,
                     figcols=2, figsize=None, **kwargs):
    """Plot correlation between features in a DataFrame.

    For each dataframe provided a subplot is generated showing a
    correlation heatmap of the features. For numeric features the method can
    be any method supported by ``pandas.DataFrame.corr``; for categorical or
    object feature types ``'cramers_v'`` must be specified. If the method does
    not match the data type an error is raised.

    The `plot_diff` parameter will also include a difference plot i.e.
    correlation difference between two dataframes. This is only used when
    two dataframes are provided.

    Parameters
    ----------
    *df: pandas.DataFrame
        Any number of DataFrame objects to plot.
    feats: str or list of str
        Feature(s) to plot. Must be column(s) in `df`.
    method: {'pearson'|'spearman'|'cramers_v'}, default='pearson'
        Correlation method. See ``pandas.DataFrame.corr`` for more details on
        ``'pearson'`` and ``'spearman'``. When ``'cramers_v'`` is specified,
        correlation is calculated using
        ``synthgauge.metrics.correlation.cramers_v``.
    plot_diff: bool, default=False
        If True and exactly two data frames are provided will also plot
        a heatmap of the absolute differences between the respective datasets'
        correlations.
    figcols: int, default=2
        Number of columns to use in figure. Only used when `feats` contains
        more than one feature.
    figsize: tuple of float, optional
        Size of figure in inches (W,H). Defaults to
        ``rcParams["figure.figsize"](default: [6.4, 4.8])``.
    **kwargs: dict, optional
        Any other keyword arguments to be passed to ``seaborn.heatmap``. For
        example ``annot=True`` will turn on cell annotations. See documentation
        for more examples.

    Returns
    -------
    matplotlib.figure.Figure
    """
    def method_cramers_v(df):
        """Compute cramers_v

        Compute pairwise cramers_v correlation for the entire dataframe.
        """
        results = np.array([])
        for x, y in product(df.columns, repeat=2):
            results = np.append(results, cramers_v(df[x], df[y]))

        size = np.sqrt(results.size).astype(int)
        results = results.reshape((size, size))

        return DataFrame(results, index=df.columns, columns=df.columns)

    if isinstance(feats, str):
        feats = [feats]

    corr_results = list()

    for d in df:
        # If no features specified use all columns
        if feats is None:
            feats = d.columns

        if method.lower() in ['pearson', 'spearman']:
            data = d[feats].select_dtypes(include='number')
            if len(data.columns) == 0:
                raise ValueError("No numeric columns available for method: "
                                 f"{method}")

            corr_results.append(data.corr(method=method).dropna(
                0, 'all').dropna(1, 'all'))

        elif method.lower() == 'cramers_v':
            data = d[feats].select_dtypes(include=['object', 'category'])
            if len(data.columns) == 0:
                raise ValueError("No categorical columns available for "
                                 f"method: {method}")

            corr_results.append(method_cramers_v(data).dropna(
                0, 'all').dropna(1, 'all'))

    # Get min and max to set consistant colourbar
    corr_values = np.array(corr_results)
    vmin = corr_values.min()
    vmax = corr_values.max()

    # For now only perform diff if 2 DataFrames are given; no more.
    # TODO: Allow all diff permutations?
    if len(corr_results) == 2 and plot_diff:
        corr_diff = np.abs(corr_results[0] - corr_results[1])
        corr_results.append(corr_diff)

    n_subplots = len(corr_results)
    ncols = 1 if n_subplots == 1 else figcols
    nrows = int(np.ceil(n_subplots/ncols))

    fig, axes = plt.subplots(nrows, ncols, figsize=figsize)

    for ax_num, (ax, corr) in enumerate(zip(np.array(axes).ravel(),
                                            corr_results)):
        if ax_num == n_subplots - 1 and plot_diff:
            sp_title = 'Correlation Difference'
            # Ignore vmin and vmax for this plot as scale will be different
            # to the others
            vmin = None
            vmax = None
        else:
            sp_title = f"DataFrame {ax_num+1} Correlation"

        sns.heatmap(corr, ax=ax, vmin=vmin, vmax=vmax, **kwargs)

        ax.set_title(sp_title)

    # Turn off axes with no data
    for ax in np.array(axes).ravel():
        if not ax.has_data():
            ax.set_visible(False)

    fig.tight_layout()
    return fig


def plot_crosstab(real, synth, x, y, x_bins='auto', y_bins='auto',
                  figsize=None, **kwargs):
    """Plot crosstab heatmap for two features.

    The two-feature crosstab calculation is performed and plotted as a heatmap.
    One heatmap is shown for the `real` data and one for the `synthetic`.
    Numeric features are discretised using the corresponding bins argument.

    Parameters
    ----------
    real: pandas.DataFrame
        DataFrame containing the real data.
    synth: pandas.DataFrame
        DataFrame containing the sythetic data.
    x: str
        Feature to plot on x axis. Must be column in `df`.
    y: str
        Feature to plot on y axis. Must be column in `df`.
    x_bins, y_bins: array_like or int or str, default='auto'
        If ``array_like`` must be sequence of bin edges. If ``int``, specifies
        the number of bins to use. If ``str``, specifies the method used to
        calculate the optimal bin width. Default= ``auto``. For all possible
        values see ``numpy.histogram_bin_edges``.
        Only used for numeric features.
    figsize: tuple of float, optional
        Size of figure in inches (W,H). Defaults to
        ``rcParams["figure.figsize"](default: [6.4, 4.8])``.
    **kwargs: dict, optional
        Any other keyword arguments to be passed to ``seaborn.heatmap``. For
        example ``annot=True`` will turn on cell annotations. See documentation
        for more examples.

    Returns
    -------
    matplotlib.figure.Figure
    """
    # Collect x and y values
    real_x, real_y = real[x], real[y]
    synth_x, synth_y = synth[x], synth[y]
    all_x = real_x.append(synth_x)
    all_y = real_y.append(synth_y)

    # Discretise numeric features
    try:
        if is_numeric_dtype(all_x):
            x_bins = np.histogram_bin_edges(all_x.dropna(), x_bins)
            real_x = cut(real_x, x_bins)
            synth_x = cut(synth_x, x_bins)

        if is_numeric_dtype(all_y):
            y_bins = np.histogram_bin_edges(all_y.dropna(), y_bins)
            real_y = cut(real_y, y_bins)
            synth_y = cut(synth_y, y_bins)
    except TypeError:
        raise TypeError('`x_bins` and `y_bins` must not be None')

    freq_real = crosstab(real_x, real_y, dropna=False)
    freq_synth = crosstab(synth_x, synth_y, dropna=False)

    fig, axes = plt.subplots(1, 2, figsize=figsize, constrained_layout=True)

    # Use same scale for real and synth
    vmin = min(freq_real.values.min(), freq_synth.values.min())
    vmax = max(freq_real.values.max(), freq_synth.values.max())
    mpbl = mpl.cm.ScalarMappable(mpl.colors.Normalize(vmin, vmax),
                                 cmap=sns.color_palette("rocket",
                                 as_cmap=True))

    sns.heatmap(freq_real, ax=axes[0], vmin=vmin, vmax=vmax, cbar=False,
                **kwargs)
    axes[0].set_title('REAL')

    sns.heatmap(freq_synth, ax=axes[1], vmin=vmin, vmax=vmax, cbar=False,
                **kwargs)
    axes[1].set_title('SYNTH')

    fig.colorbar(mpbl, ax=axes, shrink=0.8)

    return fig


def plot_qq(real, synth, feature, n_quantiles=None, figsize=None):
    """Plot a Q-Q plot.

    Generate a Quantile-Quantile plot of `feature` for the real and
    synthetic data. Q-Q plots are used to visualise two datasets
    to see if they are generated from the same distribution.

    In this case, it is used to provide some insight into the distribution
    of feature `feature` in the synthetic data versus the real data. If the
    plot is a straight line then it can be inferred that the two
    distributions are similar and therefore the synthetically generated
    data follows the same distribution as the real data.

    See `Q-Q Plot <https://en.wikipedia.org/wiki/Q-Q_plot>`_ for more
    information.

    Parameters
    ----------
    real: pandas.DataFrame
        DataFrame containing the real data.
    synth: pandas.DataFrame
        DataFrame containing the sythetic data.
    feature: str
        Feature to plot.
    n_quantiles: int, optional
        Number of quantiles to calculate. If ``None`` (the default) and
        ``real[feature]`` has the same length as ``synth[feature]`` then
        `n_quantiles` is set to the number of elements. If the lengths are
        different then n_quantiles is set to the length of ``real[feature]``.
    figsize: tuple of float, optional
        Size of figure in inches (W,H). Defaults to
        ``rcParams["figure.figsize"](default: [6.4, 4.8])``.

    Returns
    -------
    matplotlib.figure.Figure
    """
    if n_quantiles is None:
        if len(real[feature]) == len(synth[feature]):
            # Matching lengths do not require computing quantiles
            x = np.sort(real[feature])
            y = np.sort(synth[feature])
        else:
            # Determine n_quantiles from real data
            n_quantiles = len(real[feature])
            qs = np.linspace(0, 1, n_quantiles)
            x = np.quantile(real[feature], qs)
            y = np.quantile(synth[feature], qs)

    else:
        qs = np.linspace(0, 1, n_quantiles)
        x = np.quantile(real[feature], qs)
        y = np.quantile(synth[feature], qs)

    fig, ax = plt.subplots(1, 1, figsize=figsize)
    ax.scatter(x, y)
    ax.set_xlabel('real data quantiles')
    ax.set_ylabel('synth data quantiles')
    ax.set_title(f'Q-Q Plot for "{feature}"')

    fig.canvas.draw()
    if suggest_label_rotation(ax):
        ax.set_xticks(ax.get_xticks())
        ax.set_xticklabels(ax.get_xticklabels(), rotation=90)

    # Plot X = Y
    min_xy = min(x.min(), y.min())
    max_xy = max(x.max(), y.max())
    ax.plot([min_xy, max_xy], [min_xy, max_xy])

    return fig


def plot_feat_density_diff(real, synth, feats=None, feat_bins=10,
                           diff_bins=10, figsize=None):
    """Plot real and synth feature density differences.

    For each feature the feature density difference between `real` and
    `synth` is calculated using ``synthgauge.utils.feature_density_diff``.

    If a single feature is provided the plot shows the raw density
    differences for each bin.

    Where multiple feature are provided, the density differences are
    concatenated into a flattened array and a histogram plotted. The
    histogram represents the distribution of differences in densities
    across all features and bins.

    Parameters
    ----------
    real: pandas.DataFrame
        DataFrame containing the real data.
    synth: pandas.DataFrame
        DataFrame containing the sythetic data.
    feats: str or list of str, optional.
        The features that will be used to compute the densities. By
        default all features in `real` will be used.
    feat_bins: str or int, optional
        Bins to use for computing the feature densities. This value is passed
        to ``numpy.histogram_bin_edges`` so can be any value accepted by
        that function. The default setting of ``10`` uses 10 bins.
    diff_bins: str or int, optional
        Bins to use for the difference histogram. Only used when
        multiple features are provided. This value is passed to
        ``numpy.histogram_bin_edges`` so can be any value accepted by
        that function. The default setting of ``10`` uses 10 bins.

    Returns
    -------
    matplotlib.axes._subplots.AxesSubplot
        The matplotlib axes containing the plot.
    """
    if isinstance(feats, Index):
        feats = feats
    elif isinstance(feats, str):
        feats = [feats]
    else:
        feats = feats or real.columns

    if len(feats) == 1:
        diff_hist, diff_edges = feature_density_diff(real, synth, feats,
                                                     feat_bins)
        xlabel = f'{feats[0]} Binned'
        ylabel = 'Density Difference'
        title = f'Feature Density Difference for {feats[0]}'

    else:
        # TODO: option to have different bins for each feature
        diffs = [feature_density_diff(real, synth, f, feat_bins)[0]
                 for f in feats]

        diff_hist, diff_edges = np.histogram(np.concatenate(diffs),
                                             bins=diff_bins)

        xlabel = 'Difference Bins'
        ylabel = 'Count'
        title = 'Histogram of Density Differences'

    fig, ax = plt.subplots(figsize=figsize)

    # default bar width is too large so use scaled bin size
    bar_width = (diff_edges[1] - diff_edges[0]) * 0.8
    ax.bar(diff_edges[:-1], diff_hist, align='edge', width=bar_width)

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(title)

    return ax


if __name__ == '__main__':
    pass



import numpy as np
import pandas as pd
import pickle
import warnings


def load_pickle(src):
    """Load pickle file into object.

    """

    with open(src, 'rb') as pf:
        data = pickle.load(pf)
    return data


def df_combine(real, synth, feats=None, source_col_name='source',
               source_val_real='real', source_val_synth='synth'):
    """ Combine separate dataframes of real and synthetic data.

    The dataframes are concatenated along axis=0 and a source column is
    added to distinguish the real data from the sythetic data.

    Optionally, specific features can be selected.

    Parameters
    ----------
    real: pandas.DataFrame
        DataFrame containing the real, unsythesised data.
    synth: pandas.DataFrame
        DataFrame containing the synthesised data.
    feats: str or list of str, optional
        Features to combine. By default all common features are used.
    source_col_name: str, optional
        Name of the source column. This is added to the combined DataFrame
        and filled with the `source_val_real` and `source_val_synth`
        values to signify the real and synthetic data respectively.
        Defaults to "source".
    source_val_real: any, optional
        Value to use in `source_col_name` column to signify the real data.
        Any value accepted by pandas.DataFrame. Defaults to "real".
    source_val_synth: any, optional
        Value to use in `source_col_name` column to signify the sythetic
        data. Any value accepted by pandas.DataFrame. Defaults to "synth".

    Returns
    -------
    pandas.DataFrame
        The result of the concatenation of the input DataFrame objects.
    """

    feats = feats or real.columns.intersection(synth.columns)
    if isinstance(feats, str):
        feats = [feats]

    real = real[feats].copy()
    real[source_col_name] = source_val_real

    synth = synth[feats].copy()
    synth[source_col_name] = source_val_synth

    data = pd.concat([real, synth])

    return data


def df_separate(data, source_col_name, feats=None, source_val_real='real',
                source_val_synth='synth', drop_source_col=True):
    """ Separate a dataframe into real and synthetic data.

    The dataframe is split using a source column and real and synthetic
    flags. Optionally, specific features can be selected.

    Parameters
    ----------
    data: pandas.DataFrame
        DataFrame to split into real and synthetic components.
    source_col_name: str
        Name of the column used to signify real versus synthetic data.
    feats: str or list of str, optional
        Features to separate. By default all features are used.
    source_val_real: any, optional
        Value in `source_col_name` column signifying the real data. Any
        value accepted by pandas.DataFrame. Defaults to "real".
    source_val_synth: any, optional
        Value in `source_col_name` column signifying the synthetic data.
        Any value accepted by pandas.DataFrame. Defaults to "synth".
    drop_source_col: bool, optional
        If True, the default, the `source_col_name` column will be dropped
        from the output DataFrames.

    Returns
    -------
    pandas.DataFrame, pandas.DataFrame
        Two DataFrame objects containing the real data and synthetic
        data respectively.
    """
    feats = feats or data.columns
    if isinstance(feats, str):
        feats = [feats]

    real = data[data[source_col_name] == source_val_real][feats].copy()
    synth = data[data[source_col_name] == source_val_synth][feats].copy()

    if drop_source_col:
        real.drop(columns=source_col_name, inplace=True)
        synth.drop(columns=source_col_name, inplace=True)

    return real, synth


def launder(real, synth, feats=None, suffix_real='real', suffix_synth='synth'):
    """ Launder feature names and optionally filter.

    To provide clear distinction between the real and synthetic
    features, each dataframe is updated to append suffixes to the feature
    names.

    Optionally, specific features can be selected.

    Parameters
    ----------
    real: pandas.DataFrame
        DataFrame containing the real, unsythesised data
    synth: pandas.DataFrame
        DataFrame containing the synthesised data
    feats: str or list of str, optional
        Features to launder. By default all common features are used.
    suffix_real: str, optional
        Suffix to append to columns in `real`. Default is "real".
    suffix_synth: str, optional
        Suffix to append to columns in `synth`. Default is "synth".

    Returns
    -------
    pandas.DataFrame, pandas.DataFrame
        Two DataFrame objects containing the real data and synthetic
        data respectively.
    """

    feats = feats or real.columns.intersection(synth.columns)
    if isinstance(feats, str):
        feats = [feats]

    real = real[feats].copy()
    synth = synth[feats].copy()

    real.columns = [f'{c}_{suffix_real}' for c in real.columns]
    synth.columns = [f'{c}_{suffix_synth}' for c in synth.columns]

    return real, synth


def cat_encode(df, feats=None, return_all=False, convert_only=False,
               force=False):
    """ Convert object features to categories.

    Generates a new version of the input datframe with the
    specified features categorically encoded with integer labels.
    Optionally, the features can be returned as the 'category' type
    with no encoding.

    Before performing the conversion a check is made to identify any
    speficied features that are not 'object' type and therefore less
    suited to categorical encoding. A warning is raised for these features
    and they will be ignored from subsequent encoding steps unless
    `force` is `True`.

    Parameters
    ----------
    df: pandas.DataFrame
        Input dataframe to be converted.
    feats: str or list of str, optional
        Feature(s) in `df` to convert to category. By default all object
        columns are selected.
    return_all: bool, default=False
        If True all features in `df` will be returned with only those
        specified in `feats` converted. Otherwise only the converted
        features are returned. This is the default.
    convert_only: bool, default=False
        If True the features will only be converted to the 'category'
        type and NOT encoded.
    force: bool, optional
        If true, all specified features will be encoded regardless of
        data type.


    Returns
    -------
    pandas.DataFrame
        DataFrame with converted features
    dict or NoneType
        A dictionary mapping each encoded feature to its categories. If
        convert_only=True NoneType is returned.

    Warns
    -----
    UserWarning
        A warning is given if any of the specified features for
        conversion are not an 'object' type.

    """
    all_obj = df.select_dtypes(include=['object', 'category']).columns
    feats = feats or all_obj
    if isinstance(feats, str):
        feats = [feats]

    # Check for non-object type features
    non_obj = list(set(feats).difference(all_obj))
    if len(non_obj) > 0:
        warnings.warn('Selected features include non-object types:'
                      f'{non_obj} Is this intended? If so, rerun and '
                      'specify `force_encode=True`. Otherwise they will '
                      'be dropped or, if `return_all==True`, passed '
                      'through unchanged.')

    if return_all:
        out_df = df.copy()
    else:
        out_df = df[feats].copy()

    cat_dict = {} if not convert_only else None

    if force:
        enc_fts = feats
    else:
        enc_fts = list(set(feats).difference(non_obj))

    for ft in enc_fts:
        out_df[ft] = out_df[ft].astype('category')

        if not convert_only:
            cat_dict.update({ft: out_df[ft].cat.categories})
            out_df[ft] = out_df[ft].cat.codes

    return out_df, cat_dict


def feature_density_diff(real, synth, feature, bins=10):
    """Computes the difference between real and synth feature densities.

    For the specified feature the density is computed across `bins` in both
    the `real` and `synth` data sets. The per-bin difference is computed and
    returned along with the bin edges that were used.

    Prior to calculating the densities all values are converted to numeric
    via `utils.cat_encode`.

    Parameters
    ----------
    real: pandas.DataFrame
        DataFrame containing the real data.
    synth: pandas.DataFrame
        DataFrame containing the sythetic data.
    feature: str
        The feature that will be used to compute the density.
    bins: str or int, optional
        Bins to use for computing the density. This value is passed
        to `numpy.histogram_bin_edges` so can be any value accepted by
        that function. The default setting of 10 uses 10 bins.

    Returns
    -------
    hist_diff: np.ndarry
        The difference in feature density for each of the bins.
    bin_edges: np.ndarray
        The edges of the bins.

    """
    enc_real, _ = cat_encode(real, feats=feature)
    enc_synth, _ = cat_encode(synth, feats=feature)

    full_range = np.concatenate((enc_real[feature], enc_synth[feature]))
    bin_edges = np.histogram_bin_edges(full_range, bins=bins)

    real_hist, _ = np.histogram(enc_real[feature], bins=bin_edges,
                                density=True)
    synth_hist, _ = np.histogram(enc_synth[feature], bins=bin_edges,
                                 density=True)

    hist_diff = synth_hist - real_hist

    return hist_diff, bin_edges


if __name__ == '__main__':
    pass



# flake8: noqa
import pkg_resources

from . import datasets
from .evaluate import Evaluator
from . import evaluate
from . import metrics
from . import plot
from . import utils

try:
    __version__ = pkg_resources.get_distribution('synthgauge').version
except pkg_resources.DistributionNotFound:
    # Raised when package has not been installed e.g. just
    # src added to path.
    __version__ = None



import random

import numpy as np
import pandas as pd
from sklearn.datasets import make_classification


def make_blood_types_df(noise=0, proportion_nan=0, random_seed=42):
    """Create Dummy Data for examples

    This function creates dummy data for the synthgauge examples.

    Parameters
    ----------
    noise : int
        Standard deviation of the Gaussian noise to add to the data, default
        zero.
    proportion_nan: float [0,1]
        Proportion of dataset to replace with nans.
    random_seed : int
        Use for reproducibility.

    Returns
    -------
    pd.DataFrame

    Notes
    -----
    The amout of noise can be tuned to crudely simulate the creation of
    synthetic data.
    """

    X, y = make_classification(n_samples=1000,
                               n_features=5,
                               n_informative=3,
                               n_redundant=2, n_classes=4,
                               weights=[0.4, 0.3, 0.2, 0.1],
                               flip_y=0.1,
                               random_state=random_seed)

    np.random.seed(random_seed)
    random.seed(random_seed)

    mat = np.column_stack((X, 3*y))
    prop = int(mat.size * proportion_nan)
    i = [random.choice(range(mat.shape[0])) for _ in range(prop)]
    j = [random.choice(range(mat.shape[1])) for _ in range(prop)]
    mat[i, j] = np.NaN

    df = pd.DataFrame(mat + np.random.normal(scale=noise, size=(1000, 6)))

    df.columns = ['age', 'height', 'weight',
                  'hair_colour', 'eye_colour', 'blood_type']

    df.age = np.abs(round(df.age*(52/9)+44))
    df.height = np.abs(round(df.height*(52/9)+175))
    df.weight = np.abs(round(df.weight*(52/9)+80))
    df.hair_colour = (pd.cut(df.hair_colour, 4, labels=[
        'Red', 'Black', 'Brown', 'Blonde'])
        .cat.as_unordered())
    df.eye_colour = pd.cut(df.eye_colour, 3, labels=[
                           'Blue', 'Brown', 'Green']).cat.as_unordered()
    df.blood_type = pd.cut(df.blood_type, 4, labels=[
                           'O', 'A', 'B', 'AB']).cat.as_unordered()

    return df


if __name__ == '__main__':
    pass



from copy import deepcopy
from . import metrics
from .plot import (plot_correlation, plot_crosstab, plot_histograms,
                   plot_histogram3d, plot_qq)
import pandas as pd
import pickle
import warnings
from .utils import df_combine, launder


class Evaluator(object):
    """
    """

    def __init__(self, real, synth, handle_nans='drop'):
        """
        """
        common_feats = real.columns.intersection(synth.columns)
        ignore_feats = real.columns.union(synth.columns). \
            difference(common_feats)

        if len(ignore_feats) > 0:
            msg = f"Features {', '.join(ignore_feats)} are not common to " \
                "`real` and `synth` and will be ignored in further analysis"

            warnings.warn(msg)

        self.feature_names = common_feats

        # Metrics is private to apply some validation
        self.__metrics = dict()  # assign metrics and kwargs
        self.metric_results = dict()  # store results

        # Handle NaNs
        if handle_nans == 'drop':
            real.dropna(inplace=True)
            synth.dropna(inplace=True)

        self.real_data = real
        self.synth_data = synth

    def _feat_valid(self, feat):
        """Checks if feature is valid common feature
        """
        return feat in self.feature_names

    def describe_numeric(self):
        """ Summarise numerical features

        This function uses `pandas.DataFrame.describe` to calculate
        summary statistics for each numeric real and synthetic feature.

        Returns
        -------
        pandas.DataFrame
            Descriptive statistics for each numeric feature.
        """

        real, synth = launder(self.real_data, self.synth_data)

        return pd.concat([real.describe(include='number'),
                          synth.describe(include='number')],
                         axis=1).T.sort_index()

    def describe_categorical(self):
        """ Summarise categorical features

        This function uses `pandas.DataFrame.describe` to calculate
        summary statistics for each object-type real and synthetic
        feature.

        Returns
        -------
        pandas.DataFrame
            Descriptive statistics for each object-type feature.
        """

        real, synth = launder(self.real_data, self.synth_data)

        return pd.concat([real.describe(include=['category', 'object']),
                          synth.describe(include=['category', 'object'])],
                         axis=1).T.sort_index(). \
            rename(columns={'top': 'most_frequent'})

    def add_metric(self, metric_name, metric_alias=None, **metric_kwargs):
        """ Add a metric to the evaluator

        Metrics and their arguments are recorded to be run at
        a later time. This allows metric customisation but ensures
        that the same metric configuration is applied consistently i.e.
        once added the parameters do not require resupplying for each
        execution of the metric. Supplying a metric alias allows the same
        metric to be used multiple times with different parameters.

        Note that self.real_data and self.synth_data will be passed
        automatically to metrics that expect these arguments. They should
        not be declared in `metric_kwargs`.

        Parameters
        ----------
        metric_name: str
            Name of the metric. Must match function name in metrics.
        metric_alias: str, optional
            Alias to be given to this use of the metric. Allows the same metric
            to be used multiple times with different parameters within the same
            evaluator instance.

        **metric_kwargs : dict, optional
            Extra arguments to `metric_name`: refer to each metric
            documentation for a list of all possible arguments.
        """
        try:
            getattr(metrics, metric_name)
            if metric_alias is None:
                self.__metrics.update({metric_name: metric_kwargs})
            else:
                metric_kwargs.update({'metric_name': metric_name})
                self.__metrics.update({metric_alias: metric_kwargs})
        except AttributeError:
            raise NotImplementedError(f"Metric '{metric_name}' is not "
                                      "implemented")

    def add_custom_metric(self, metric_name, metric_func, **metric_kwargs):
        """Add a custom metric to the Evaluator object.

        To enhance customisability, this function allows users to add metrics
        from outwith SynthGauge to the Evaluator.

        A custom metric is a function that accepts the real and synthetic
        dataframes as the first and second positional arguments respectively.
        Any other parameters must be defined as keyword arguments. The metric
        function can return a value of any desired type although scalar numeric
        values are recommended, or namedtuples when there are multiple
        outputs.

        Parameters
        ----------
        metric_name: str
            Name of the metric. This is what will appear in the results table.
        metric_func: function
            Function to be called during the evaluation step. The first two
            arguments will be ``self.real`` and ``self.synth``.
        **metric_kwargs : dict, optional
            Extra arguments to be passed to `metric_func` during evaluation.
        """
        metric_kwargs.update({'metric_func': metric_func})
        self.__metrics.update({metric_name: metric_kwargs})

    def copy_metrics(self, Other):
        """Copy metrics from another Evaluator object

        To facilitate consistent comparisons of different synthetic datasets,
        this function copies the metrics dictionary from another evaluator
        object.

        Parameters
        ----------
        Other: Evaluator
            The other evaluator object from which the metrics dictionary will
            be copied.
        """
        if not isinstance(Other, Evaluator):
            raise TypeError("Other must be of class Evaluator")
        self.__metrics = deepcopy(Other.metrics)

    def save_metrics(self, filename):
        """Save metrics to disk

        Save the Evaluator's current metrics to a pickle file.

        Parameters
        ----------
        filename:
            Path to pickle file to save the metrics.
        """
        with open(filename, 'wb') as f:
            pickle.dump(self.metrics, f)

    def load_metrics(self, filename, overwrite=False):
        """Load metrics from disk

        Update or overwrite the Evaluator's current metrics from a pickle
        file.

        Parameters
        ----------
        filename: str
            Path to metrics pickle file.
        overwrite: bool, optional
            If True, all current metrics will be replaced with the loaded
            metrics. Default is False which will update the current metric
            dictionary with the loaded metrics.
        """
        with open(filename, 'rb') as f:
            new_metrics = pickle.load(f)

        invalid_metrics = []
        for k, v in new_metrics.items():
            if getattr(metrics, v['metric_name'], None) is None:
                invalid_metrics.append(k)

        if len(invalid_metrics) > 0:
            invalid_str = ', '.join(invalid_metrics)
            raise ValueError(f'Invalid metrics encountered in: {invalid_str}.')
        elif overwrite:
            self.__metrics = new_metrics
        else:
            self.__metrics.update(new_metrics)

    @property
    def metrics(self):
        """Return __metrics"""
        return self.__metrics

    @property
    def combined_data(self):
        """Return combined real and synthetic data"""
        return df_combine(self.real_data, self.synth_data)

    def drop_metric(self, metric):
        """Drops the metric from the evaluator

        The metric `metric` will be removed from the metric
        catalogue for the evaluator.

        Note: To update the metric parameters see add_metric.

        Parameters
        ----------
        metric: str
            Name or alias of the metric to remove from the metrics stored for
            this evaluator instance.
        """
        try:
            del self.__metrics[metric]
        except KeyError:
            pass

    def evaluate(self, as_df=False):
        """ Compute metrics for real and synth data

        Runs through the given metrics in self.__metrics and executes
        each with corresponding arguments. The results are returned as
        either a dictionary or DataFrame.

        Results are also silenty stored as a dictionary in
        `self.metric_results`.

        Parameters
        ----------
        as_df: bool, optional
            If True the results will be returned as a pandas DataFrame,
            otherwise a dictionary is returned. Default is False.

        Returns
        -------
        pandas.DataFrame
            If `as_df` is True a DataFrame is returned. The rows
            represent metric names and the columns their values.
        dict
            If `as_df` is False (the default) a dictionary is returned.
            The keys represent metric names and the values metric values.
        """
        results = dict.fromkeys(self.__metrics.keys())

        metrics_copy = deepcopy(self.__metrics)
        for metric, kwargs in metrics_copy.items():
            if 'metric_name' in kwargs.keys():
                metric_name = kwargs.pop('metric_name')
            else:
                metric_name = metric
            if metric_name in metrics.__dict__.keys():
                metric_func = getattr(metrics, metric_name)
            else:
                metric_func = kwargs.pop('metric_func')
            results[metric] = metric_func(self.real_data, self.synth_data,
                                          **kwargs)

        self.metric_results = {k: v for k, v in results.items()}

        if as_df:
            tidy_results = {}
            for k, v in self.metric_results.items():
                try:
                    for vk, vv in v._asdict().items():
                        tidy_results[k+'-'+vk] = vv
                except AttributeError:
                    tidy_results[k] = v

            return pd.DataFrame(tidy_results, index=['value']).T

        else:
            return results

    def plot_histograms(self, figcols=2, figsize=None):
        """Plot grid of feature distributions

        Convenience wrapper for plot.plot_histograms. This function uses the
        combined real and synthetic data sets and groups by 'source'.
        """
        return plot_histograms(self.combined_data, feats=self.feature_names,
                               groupby='source', figcols=figcols,
                               figsize=figsize)

    def plot_histogram3d(self, data, x, y, x_bins='auto', y_bins='auto',
                         figsize=None):
        """Plot 3D histogram

        Convenience wrapper for plot.plot_histogram3d.

        Parameters
        ----------
        data: {"real"|"synth"|"combined"}:
            Dataframe to pass to plotting function. Either "real" to pass
            `E.real_data`, "synth" to pass E.synth_data or "combined" to
            pass E.combined_data.
        """

        return plot_histogram3d(getattr(self, f'{data}_data'), x=x, y=y,
                                x_bins=x_bins, y_bins=y_bins, figsize=figsize)

    def plot_correlation(self, feats=None, method='pearson', figcols=2,
                         figsize=None, **kwargs):
        """Plot correlation heatmaps

        Plot correlation heatmaps for `self.real_data`, `self.synth_data`.
        See plot.plot_correlation for details.

        """
        return plot_correlation(self.real_data, self.synth_data, feats=feats,
                                method=method, plot_diff=True, figcols=figcols,
                                figsize=figsize, **kwargs)

    def plot_crosstab(self, x, y, figsize=None, **kwargs):
        """Plot pairwise crosstabulation

        Convenience wrapper for plot.plot_crosstab. Automatically sets `real`
        and `synth` parameters to the correspondin data in the Evaluator.
        """
        return plot_crosstab(self.real_data, self.synth_data, x=x, y=y,
                             figsize=figsize, **kwargs)

    def plot_qq(self, feature, n_quantiles=None, figsize=None):
        """Plot Q-Q plot

        Plot Quantile-Quantile plot. See plot.plot_qq for details.

        Parameters
        ----------
        feature: str
            Feature to plot.
        **kwargs: dict, optional
            Keyword arguments to pass to plot.plot_qq.
        """
        return plot_qq(self.real_data, self.synth_data, feature,
                       n_quantiles, figsize)


if __name__ == '__main__':
    pass



from .classification import classification_comparison
from .cluster import multi_clustered_MSD
from .correlation import correlation_MSD, correlation_ratio_MSE, cramers_v_MSE
from .privacy import TCAP, min_NN_dist, sample_overlap_score
from .propensity import propensity_metrics
from .univariate_distance import (
     feature_density_diff_mae, jensen_shannon_distance,
     jensen_shannon_divergence, kolmogorov_smirnov, kruskal_wallis,
     kullback_leibler, mann_whitney, wasserstein, wilcoxon
)

__all__ = ['kolmogorov_smirnov', 'wasserstein', 'jensen_shannon_distance',
           'feature_density_diff_mae', 'kullback_leibler',
           'jensen_shannon_divergence', 'kruskal_wallis', 'mann_whitney',
           'wilcoxon', 'correlation_MSD', 'cramers_v_MSE',
           'correlation_ratio_MSE', 'propensity_metrics',
           'multi_clustered_MSD', 'classification_comparison', 'TCAP',
           'min_NN_dist', 'sample_overlap_score']
